import re
from math import ceil
from itertools import islice
from tqdm.auto import tqdm
from elastic_wikidata.sparql_helpers import run_query, paginate_sparql_query
from elastic_wikidata.http import generate_user_agent


def url_to_qid(url: str) -> str:
    """
    Maps Wikidata URL of an entity to QID e.g. http://www.wikidata.org/entity/Q7187777 -> Q7187777.
    """

    return re.findall(r"(Q\d+)", url)[0]


def get_entities_from_query(query, page_size=None, limit=None) -> list:
    """
    Get a list of entities from a query. Optionally:
        paginate the query using page_size
        limit the total number of entities returned using limit

    Returns list of entities in form (Qd+).
    """

    if page_size:
        pages = paginate_sparql_query(query, page_size=page_size)
    else:
        pages = [query]

    if limit:
        page_limit = ceil(limit / page_size)
        pages = islice(pages, page_limit)

    all_entities = []

    for query in tqdm(pages, total=(page_limit or None)):
        res = run_query(query)
        var = res["head"]["vars"][0]
        entities = [url_to_qid(x[var]["value"]) for x in res["results"]["bindings"]]
        all_entities += entities

        # stop when page of query returns fewer items than the page size
        if len(entities) < page_size:
            break

    return all_entities



import json
from itertools import islice
from tqdm.auto import tqdm
from elasticsearch import Elasticsearch
from elasticsearch.helpers import parallel_bulk
from typing import Union
from elastic_wikidata.wd_entities import (
    get_entities,
    wiki_property_check,
    simplify_wbgetentities_result,
)


class processDump:
    def __init__(
        self, dump: Union[str, list], es_credentials: dict, index_name: str, **kwargs
    ):
        self.config = {
            "chunk_size": 1000,
            "queue_size": 8,
        }

        self.es_credentials = es_credentials

        if isinstance(dump, str):
            self.dump_path = dump
            self.entities = None
        elif isinstance(dump, list):
            self.entities = dump
            self.dump_path = None
        else:
            raise ValueError(
                "dump must either be path to JSON dump or Python list of entities"
            )

        self.index_name = index_name

        # process kwargs/set defaults
        self.disable_refresh_on_index = kwargs["disable_refresh_on_index"]

        if "doc_limit" in kwargs:
            self.doc_limit = kwargs["doc_limit"]
        else:
            self.doc_limit = None

        self.wiki_options = {}

        if "lang" in kwargs:
            self.wiki_options["lang"] = kwargs["lang"]
        else:
            self.wiki_options["lang"] = "en"

        if "user_agent_contact" in kwargs:
            self.user_agent_contact = kwargs["user_agent_contact"]
        else:
            self.user_agent_contact = None

        if "properties" in kwargs:
            if isinstance(kwargs["properties"], str) and wiki_property_check(
                kwargs["properties"]
            ):
                self.wiki_options["properties"] = [
                    kwargs["properties"].upper()
                ]  # [P31], not [p31]
            elif isinstance(kwargs["properties"], list):
                self.wiki_options["properties"] = [
                    item.upper()
                    for item in kwargs["properties"]
                    if wiki_property_check(item)
                ]
        else:
            self.wiki_options["properties"] = ["P31"]

    def start_elasticsearch(self):
        """
        Creates an Elasticsearch index. If SEARCH_CLUSTER, ELASTICSEARCH_USER & ELASTICSEARCH_PASSWORD
        are specified in config it uses those, otherwise uses a locally running Elasticsearch instance.
        """

        if "ELASTICSEARCH_CLUSTER" in self.es_credentials:
            print(
                f"Connecting to Elasticsearch at {self.es_credentials['ELASTICSEARCH_CLUSTER']}"
            )
            self.es = Elasticsearch(
                [self.es_credentials["ELASTICSEARCH_CLUSTER"]],
                http_auth=(
                    self.es_credentials["ELASTICSEARCH_USER"],
                    self.es_credentials["ELASTICSEARCH_PASSWORD"],
                ),
            )
        else:
            # run on localhost
            print("Connecting to Elasticsearch on localhost")
            self.es = Elasticsearch()

        mappings = {
            "mappings": {
                "properties": {
                    "labels": {"type": "text", "copy_to": "labels_aliases"},
                    "aliases": {"type": "text", "copy_to": "labels_aliases"},
                    "labels_aliases": {"type": "text", "store": "true"},
                }
            }
        }

        self.es.indices.create(index=self.index_name, ignore=400, body=mappings)

        if self.disable_refresh_on_index:
            print(
                "Temporary disabling refresh for the index. Will reset refresh interval to the default (1s) after load is complete."
            )
            self.es.indices.put_settings({"index": {"refresh_interval": -1}})

    def dump_to_es(self):
        print("Indexing documents...")
        successes = 0
        errors = []

        # if dump_path, use generator that passes
        if self.dump_path:
            action_generator = self.generate_actions_from_dump()
        elif self.entities:
            action_generator = self.generate_actions_from_entities()

        for ok, action in tqdm(
            parallel_bulk(
                client=self.es,
                index=self.index_name,
                actions=action_generator,
                chunk_size=self.config["chunk_size"],
                queue_size=self.config["queue_size"],
            ),
        ):
            if not ok:
                print(action)
                errors.append(action)
            successes += ok

        if self.disable_refresh_on_index:
            # reset back to default
            print("Refresh interval set back to default of 1s.")
            self.es.indices.put_settings({"index": {"refresh_interval": "1s"}})

    def process_doc(self, doc: dict) -> dict:
        """
        Processes a single document from the JSON dump, returning a filtered version of that document. 
        """

        lang = self.wiki_options["lang"]
        properties = self.wiki_options["properties"]

        return simplify_wbgetentities_result(doc, lang, properties)

    def generate_actions_from_dump(self):
        """
        Generator to yield a processed document from the Wikidata JSON dump. 
        Each line of the Wikidata JSON dump is a separate document. 
        """
        with open(self.dump_path) as f:
            objects = (json.loads(line) for line in f)

            # optionally limit number that are loaded
            if self.doc_limit is not None:
                objects = islice(objects, self.doc_limit)

            for item in objects:
                doc = self.process_doc(item)

                yield doc

    def generate_actions_from_entities(self):
        """
        Generator to yield processed document from list of entities. Calls are made to 
        wbgetentities API with page size of 50 to retrieve documents.
        """

        json_generator = get_entities.result_generator(
            self.entities, lang=self.wiki_options["lang"]
        )

        for page in json_generator:
            for item in page:
                yield self.process_doc(item)



from elastic_wikidata import dump_to_es, sparql_to_es
from elastic_wikidata.config import runtime_config
import click
from configparser import ConfigParser


@click.command()
@click.argument("source", nargs=1)
@click.option("--path", "-p", type=click.Path(exists=True))
@click.option(
    "--cluster", envvar="ELASTICSEARCH_CLUSTER", help="Elasticsearch cluster URL"
)
@click.option("--user", envvar="ELASTICSEARCH_USER", help="Elasticsearch username")
@click.option(
    "--password", envvar="ELASTICSEARCH_PASSWORD", help="Elasticsearch password"
)
@click.option(
    "--agent_contact",
    "-contact",
    envvar="WIKIMEDIA_AGENT_CONTACT",
    help="(optional) Contact details to add to the User Agent header for Wikidata requests",
    default=None,
)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    help="Path to .ini file containing Elasticsearch credentials",
)
@click.option(
    "--index",
    "-i",
    prompt="Elasticsearch index",
    help="Name of Elasticsearch index to load into",
)
@click.option(
    "--limit", "-l", type=int, help="(optional) Limit the number of entities loaded in"
)
@click.option("--page_size", type=int, help="Page size for SPARQL query.", default=100)
@click.option(
    "--language", "-lang", type=str, help="Language (Wikimedia language code)"
)
@click.option(
    "--properties",
    "-prop",
    type=str,
    help="One or more Wikidata property e.g. p31 or p31,p21. Not case-sensitive",
)
@click.option(
    "--timeout",
    "-t",
    type=int,
    help="Timeout for Wikidata requests (seconds)",
    default=6,
)
@click.option(
    "--disable_refresh/--no_disable_refresh",
    "-dr/-ndr",
    help="Whether to disable CPU-intensive refresh on load. Defaults to True. Recommended to leave this on for low-resource machines or large datasets.",
    default=True,
)
def main(
    source,
    path,
    cluster,
    user,
    password,
    agent_contact,
    config,
    index,
    limit,
    page_size,
    language,
    properties,
    timeout,
    disable_refresh,
):

    # get elasticsearch credentials
    if config:
        # read .ini file
        parser = ConfigParser()
        parser.optionxform = str  # make option names case sensitive
        parser.read(config)
        es_credentials = parser._sections["ELASTIC"]
        check_es_credentials(es_credentials)

        runtime_config.add_item(
            {
                "user_agent_contact": parser._sections["HTTP"].get(
                    "CONTACT_DETAILS", None
                )
            }
        )
    else:
        # check environment variables/flags
        es_credentials = {}

        if cluster:
            es_credentials["ELASTICSEARCH_CLUSTER"] = cluster
        if user:
            es_credentials["ELASTICSEARCH_USER"] = user
        if password:
            es_credentials["ELASTICSEARCH_PASSWORD"] = password

        check_es_credentials(es_credentials)

        runtime_config.add_item({"user_agent_contact": agent_contact})

    runtime_config.add_item({"http_timeout": timeout})

    # global flag for all functions that the module is being run through the CLI
    runtime_config.add_item({"cli": True})

    # set kwargs
    kwargs = {}
    if language:
        kwargs["lang"] = language
    if properties:
        kwargs["properties"] = properties.split(",")
    if disable_refresh:
        kwargs["disable_refresh_on_index"] = disable_refresh

    # run job
    if source == "dump":
        load_from_dump(path, es_credentials, index, limit, **kwargs)
    elif source == "query":
        load_from_sparql(path, es_credentials, index, limit, page_size, **kwargs)
    else:
        raise ValueError(f"Argument {source} must be either dump or sparql")


def load_from_dump(path, es_credentials, index, limit, **kwargs):
    if not kwargs:
        kwargs = {}
    if limit:
        kwargs["doc_limit"] = limit

    # limit is used when dumping JSON to Elasticsearch
    d = dump_to_es.processDump(
        dump=path, es_credentials=es_credentials, index_name=index, **kwargs
    )
    d.start_elasticsearch()
    d.dump_to_es()


def load_from_sparql(path, es_credentials, index, limit, page_size=100, **kwargs):
    if not kwargs:
        kwargs = {}

    with open(path, "r") as f:
        query = f.read()

    # limit is used when getting list of entities
    print(f"Getting entities from SPARQL query")
    entity_list = sparql_to_es.get_entities_from_query(
        query, page_size=100, limit=limit
    )

    print(
        f"Retrieving information from wbgetentities API and pushing to ES index {index}"
    )
    d = dump_to_es.processDump(
        dump=entity_list, es_credentials=es_credentials, index_name=index, **kwargs
    )
    d.start_elasticsearch()
    d.dump_to_es()


def check_es_credentials(credentials: dict):
    credentials_present = set(credentials.keys())
    credentials_required = {
        "ELASTICSEARCH_CLUSTER",
        "ELASTICSEARCH_USER",
        "ELASTICSEARCH_PASSWORD",
    }
    missing_credentials = credentials_required - credentials_present

    if len(missing_credentials) > 0:
        raise ValueError(f"Missing Elasticsearch credentials: {missing_credentials}")


if __name__ == "__main__":
    # main(
    #     source='dump',
    #     path="../wikidata/all_no_articles.ndjson",
    #     properties="p31,p279",
    #     config="./config.ini",
    #     index='wikidump2',
    #     cluster=None,
    #     user=None,
    #     password=None,
    #     agent_contact=False,
    #     limit=None,
    #     page_size=100,
    #     language='en',
    #     timeout=6,
    #     disable_refresh=True
    # )
    main()



import requests
import sys
from urllib.parse import quote
from elastic_wikidata import __version__ as ew_version
from elastic_wikidata.config import runtime_config


def generate_user_agent():
    """
    Generates user agent string according to Wikidata User Agent Guidelines (https://meta.wikimedia.org/wiki/User-Agent_policy).
    Uses contact information from `runtime_config.get('user_agent_contact')`.

    Returns:
        str: user agent string
    """
    v_params = {
        "python": "Python/" + ".".join(str(i) for i in sys.version_info),
        "http_backend": "requests/" + requests.__version__,
        "ew": "Elastic Wikidata bot/" + ew_version,
    }

    contact_information = runtime_config.get("user_agent_contact")

    if contact_information is not None:
        contact_information = " ".join(
            [process_user_agent_username(i) for i in contact_information.split(" ")]
        )
        return f"{v_params['ew']} ({contact_information}) {v_params['http_backend']} {v_params['python']}"
    else:
        if runtime_config.get("cli"):
            print(
                "WARNING: please consider adding contact information through config.ini or the -contact flag to improve the User Agent header for Wikidata requests."
            )
        return f"{v_params['ew']} {v_params['http_backend']} {v_params['python']}"


def process_user_agent_username(username=None):
    """
    **Credit to [pywikibot](https://www.mediawiki.org/wiki/Manual:Pywikibot)**

    Reduce username to a representation permitted in HTTP headers.

    To achieve that, this function:
    1) replaces spaces (' ') with '_'
    2) encodes the username as 'utf-8' and if the username is not ASCII
    3) URL encodes the username if it is not ASCII, or contains '%'
    """
    if not username:
        return ""

    username = username.replace(" ", "_")  # Avoid spaces or %20.
    try:
        username.encode("ascii")  # just test, but not actually use it
    except UnicodeEncodeError:
        username = quote(username.encode("utf-8"))
    else:
        # % is legal in the default $wgLegalTitleChars
        # This is so that ops know the real pywikibot will not
        # allow a useragent in the username to allow through a hand-coded
        # percent-encoded value.
        if "%" in username:
            username = quote(username)
    return username



import requests
from tqdm.auto import tqdm
from typing import List, Union
from math import ceil
import re
from elastic_wikidata.http import generate_user_agent
from elastic_wikidata.config import runtime_config


class get_entities:
    def __init__(self):
        """
        One instance of this class per list of qcodes. The JSON response for a list of qcodes is made to Wikidata on 
        creation of a class instance. 

        Args:
            qcodes (str/list): Wikidata qcode or list of qcodes/
            lang (str, optional): Defaults to 'en'.
            page_limit (int): page limit for Wikidata API. Usually 50, can reach 500. 
        """
        self.endpoint = (
            "http://www.wikidata.org/w/api.php?action=wbgetentities&format=json"
        )

        self.properties = ["labels", "aliases", "claims", "descriptions"]

    @staticmethod
    def _param_join(params: List[str]) -> str:
        """
        Joins list of parameters for the URL. ['a', 'b'] -> "a%7Cb"

        Args:
            params (list): list of parameters (strings)

        Returns:
            str
        """

        return "%7C".join(params) if len(params) > 1 else params[0]

    @classmethod
    def get_all_results(
        self, qcodes, lang="en", page_limit=50, timeout: int = None
    ) -> list:
        """
        Get response through the `wbgetentities` API. 

        Returns:
            list: each item is a the response for an entity
        """

        results = self().result_generator(qcodes, lang, page_limit, timeout)

        all_results = []

        print(f"Getting {len(qcodes)} wikidata documents in pages of {page_limit}")

        for res in tqdm(results, total=ceil(len(qcodes) / page_limit)):
            all_results += res

        return all_results

    @classmethod
    def result_generator(
        self, qcodes, lang="en", page_limit=50, timeout: int = None
    ) -> list:
        """
        Get response through the `wbgetentities` API. Yields `page_limit` entities at a time.

        Returns:
            list: each item is a the response for an entity
        """

        if isinstance(qcodes, str):
            qcodes = [qcodes]

        qcodes_paginated = [
            qcodes[i : i + page_limit] for i in range(0, len(qcodes), page_limit)
        ]

        headers = {"User-Agent": generate_user_agent()}

        if timeout is None:
            timeout = runtime_config.get("http_timeout")

        with requests.Session() as s:
            for page in qcodes_paginated:
                url = f"http://www.wikidata.org/w/api.php?action=wbgetentities&format=json&ids={self._param_join(page)}&props={self._param_join(self().properties)}&languages={lang}&languagefallback=1&formatversion=2"
                response = s.get(url, headers=headers, timeout=timeout).json()
                yield [v for _, v in response["entities"].items()]

    def get_labels(self, qcodes, lang="en", page_limit=50, timeout: int = None) -> dict:
        """
        Get labels from Wikidata qcodes. If the item associated with a qcode has no label, its value
        in the dictionary is an empty string.

        Returns:
            dict: {qid1: label1, qid2: label2, ...}
        """

        qid_label_mapping = dict()
        qcodes = list(set(qcodes))

        docs = self.get_all_results(qcodes, lang, page_limit, timeout)

        for doc in docs:
            qid_label_mapping[doc["id"]] = doc["labels"].get(lang, {}).get("value", "")

        return qid_label_mapping


def simplify_wbgetentities_result(
    doc: Union[dict, List[dict]], lang: str, properties: list
) -> Union[dict, List[dict]]:
    """
    Processes a single document or set of documents from the JSON result of wbgetentities, returning a simplified version of that document. 

    Args:
        doc (Union[dict, List[dict]]): JSON result from Wikidata wbgetentities API
        lang (str): Wikimedia language code
        properties (list): list of Wikidata properties

    Returns:
        Union[dict, List[dict]]: dict if single record passed in; list if multiple records
    """

    # if list of dicts, run this function for each dict
    if isinstance(doc, list) and isinstance(doc[0], dict):
        return [simplify_wbgetentities_result(item, lang, properties) for item in doc]

    wd_type_mapping = {
        "wikibase-entityid": "id",
        "time": "time",
        "monolingualtext": "text",
    }

    newdoc = {"id": doc["id"]}

    # add label(s)
    if lang in doc["labels"]:
        newdoc["labels"] = doc["labels"][lang]["value"]

    # add descriptions(s)
    if lang in doc["descriptions"]:
        newdoc["descriptions"] = doc["descriptions"][lang]["value"]

    # add aliases
    if (len(doc["aliases"]) > 0) and (lang in doc["aliases"]):
        newdoc["aliases"] = [i["value"] for i in doc["aliases"][lang]]
    else:
        newdoc["aliases"] = []

    # add claims (property values)
    newdoc["claims"] = {}

    for p in properties:
        if p in doc["claims"]:
            claims = []
            for i in doc["claims"][p]:
                try:
                    value_type = i["mainsnak"]["datavalue"]["type"]
                    if value_type == "string":
                        claims.append(i["mainsnak"]["datavalue"]["value"])
                    else:
                        value_name = wd_type_mapping[value_type]
                        claims.append(i["mainsnak"]["datavalue"]["value"][value_name])
                except KeyError:
                    pass

            newdoc["claims"][p] = claims

    return newdoc


def wiki_property_check(p):
    if len(re.findall(r"(p\d+)", p.lower())) == 1:
        return True
    else:
        print(f"WARNING: property {p} is not a valid Wikidata property")
        return False



class RuntimeConfig:
    def __init__(self):
        self.items = {}

    def add_item(self, item: dict):
        """
        Add an item to the runtime config
        """

        self.items.update(item)

    def get(self, key: str):
        """
        Get specific item from config. Returns None if key doesn't exist.
        """

        return self.items.get(key, None)

    def get_all(self) -> dict:
        """
        Return all items from runtime config
        """

        return self.items


runtime_config = RuntimeConfig()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="elastic-wikidata",
    version="0.3.4",
    author="Science Museum Group",
    description="elastic-wikidata",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheScienceMuseum/elastic-wikidata",
    download_url="https://github.com/TheScienceMuseum/elastic-wikidata/archive/v0.3.2.tar.gz",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "click==7.1.2",
        "elasticsearch==7.8.1",
        "SPARQLWrapper==1.8.5",
        "tqdm>=4.48.2",
        "requests==2.24.0",
    ],
    py_modules=["cli", "elastic_wikidata"],
    packages=["elastic_wikidata"],
    entry_points="""
    [console_scripts]
    ew=cli:main
    """,
)



from SPARQLWrapper import SPARQLWrapper, JSON
import urllib
import time
from elastic_wikidata.http import generate_user_agent


def run_query(query: str, endpoint_url="https://query.wikidata.org/sparql") -> dict:
    """
    Run a SPARQL query against the Wikidata endpoint. Obeys retry-after headers for sensible bulk querying.

    Args:
        query (str): SPARQL query
        endpoint_url (optional)

    Returns:
        query_result (dict): the JSON result of the query as a dict
    """

    user_agent = generate_user_agent()

    sparql = SPARQLWrapper(endpoint_url, agent=user_agent)
    sparql.setQuery(query)
    sparql.setMethod("POST")
    sparql.setReturnFormat(JSON)

    try:
        return sparql.query().convert()
    except urllib.error.HTTPError as e:
        if e.code == 429:
            if isinstance(e.headers.get("retry-after", None), int):
                time.sleep(e.headers["retry-after"])
            else:
                time.sleep(10)
            return run_query(query, endpoint_url)
        raise


def paginate_sparql_query(query: str, page_size: int):
    """
    Paginates a SELECT query, returning a generator which yields paginated queries. 
    """

    # check query
    if "select" not in query.lower():
        raise ValueError("Must be a SELECT query")

    if "order by" not in query.lower():
        print(
            "WARNING: no ORDER BY logic in the SPARQL query. This could result in duplicate or missing entities."
        )

    # paginate
    i = 0
    while True:
        yield f"""{query}
        LIMIT {page_size}
        OFFSET {i*page_size}
        """
        i += 1
